
import ply.lex as lex
import ply.yacc as yacc

from model import Predicate, Clause, Expression, Optional, Reference, Stop, String, Variable, Module

class Lexer:
    tokens = (
       'BAR',
       'CLOSE_ANGLE',
       'CLOSE_BRACKET',
       'COLON',
       'NAME',
       'OPEN_ANGLE',
       'OPEN_BRACKET',
       'STOP',
       'STRING'
    )

    t_BAR =             r'\|'
    t_CLOSE_ANGLE =     r'\>'
    t_CLOSE_BRACKET =   r'\]'
    t_COLON =           r'\:'
    t_OPEN_ANGLE =      r'\<'
    t_OPEN_BRACKET =    r'\['
    t_STOP =            r'.'
    
    t_ignore  = ' \t'
    
    singleton = None
    
    @classmethod
    def the_lexer(cls):
        if not cls.singleton:
            cls.singleton = Lexer()
        return cls.singleton.lexer()
    
    @classmethod
    def tokenize(cls, data):
        lexer = cls.the_lexer()
        lexer.input(data)
        result = []
        while True:
            token = lexer.token()
            if not token: break
            result.append(token)
        return result
        
    def __init__(self):
        self.lex = None
    
    def lexer(self):
        if not self.lex:
            self.lex = lex.lex(object=self) 
        return self.lex
        
    def t_NAME(self, t):
        r'[a-zA-Z][a-zA-Z0-9\-\_]*'
        return t
    
    def t_STRING(self, t):
        r'\'(\\.|[^\'])*\''
        return t
    
    def t_error(self, t):
        raise Exception("Illegal character '%s'" % t.value[0])
    
def p_predicate(p):
    '''
       predicate : NAME COLON expression-list STOP
                 | NAME COLON expression-list
    '''
    if len(p) == 5:
        p[3].append(Stop())
    p[0] = Predicate(p[1], [Clause(p[3])])

def p_expression_list(p):
    '''
       expression-list : expression
                       | expression-list expression
    '''
    if len(p) == 3:
        p[0] = p[1]
        p[0].append(p[2])
    else:
        p[0] = [p[1]]

def p_expression(p):
    '''
       expression : term 
                  | term BAR expression
    '''
    if len(p) == 4:
        p[0] = Expression(p[1], alternate_expression=p[3])
    else:
        p[0] = Expression(p[1])

def p_reference(p):
    '''
        reference : COLON NAME
    '''
    p[0] = Reference(p[2])
    
def p_term(p):
    '''
       term : NAME
            | STRING
            | reference
            | variable
            | optional
    '''
    token = p.slice[1]
    if token.type == 'NAME':
        p[0] = String(token.value)
    elif token.type == 'STRING':
        p[0] = String(token.value[1:-1])
    else:
        p[0] = p[1]

def p_variable(p):
    '''
       variable : OPEN_ANGLE NAME CLOSE_ANGLE
    '''
    p[0] = Variable(p[2])

def p_optional(p):
    '''
       optional : OPEN_BRACKET expression-list CLOSE_BRACKET
    '''
    p[0] = Optional(p[2])

tokens = Lexer.tokens
parser = yacc.yacc()

def parse(data):
    result = parser.parse(data, lexer=Lexer.the_lexer())
    return result
    
def predicates(data):
    result = Module()
    for line in data.split('\n'):
        line = line.strip()
        if line:
            pattern = parse(line)
            result.add(pattern)
    result.resolve()
    return result