
class Module:
    def __init__(self):
        self.predicates = {}
    
    def __str__(self):
        return '\n'.join([str(predicate) for predicate in self.predicates.values()])
    
    def add(self, predicate):
        if predicate.name in self.predicates:
            self.predicates[predicate.name].merge(predicate)
        else:
            self.predicates[predicate.name] = predicate
    
    def get_children(self):
        return self.predicates.values()
    
    @classmethod
    def is_disjunction(cls):
        return True
    
    @classmethod
    def is_leaf(cls):
        return False
    
    def resolve(self):
        def resolve_named_term(term, predicates):
            if isinstance(term, Reference):
                if not term.value in predicates:
                    raise Exception("No resolution for name: '%s'" % term.name)
                term.reference = predicates[term.value]
        self.walk(resolve_named_term, self.predicates)
    
    def walk(self, visit, carry):
        for predicate in self.predicates.values():
            predicate.walk(visit, carry)
                
class Predicate:
    def __init__(self, name, clauses=[]):
        self.name = name
        self.clauses = clauses
    
    def get_children(self):
        return self.clauses
    
    @classmethod
    def is_disjunction(cls):
        return True
    
    @classmethod
    def is_leaf(cls):
        return False
    
    def merge(self, predicate):
        if self.name == predicate.name:
            self.clauses.extend(predicate.clauses)
        else:
            raise Exception("Predicate mismatch %s/%s" % (self.name, predicate.name))

    def __str__(self):
        return '\n'.join([self.name + ': ' + str(clause) for clause in self.clauses])
    
    def walk(self, visit, carry):
        visit(self, carry)
        for clause in self.clauses:
            clause.walk(visit, carry)

class Clause:
    def __init__(self, expressions):
        self.expressions = expressions
    
    def __str__(self):
        return Expression.str_list(self.expressions)
    
    def get_children(self):
        return self.expressions
    
    @classmethod
    def is_disjunction(cls):
        return False
    
    @classmethod
    def is_leaf(cls):
        return False
    
    def walk(self, visit, carry):
        visit(self, carry)
        for expression in self.expressions:
            expression.walk(visit, carry)

class Expression:
    def __init__(self, term, alternate_expression=None):
        self.term = term
        self.alternate_expression = alternate_expression
    
    def __str__(self):
        if self.alternate_expression:
            return str(self.term) + ' | ' + str(self.alternate_expression)
        else:
            return str(self.term)
    
    def get_children(self):
        if self.alternate_expression:
            return [self.alternate_expression, self.term]
        else:
            return [self.term]
    
    @classmethod
    def is_disjunction(cls):
        return True
    
    @classmethod
    def is_leaf(cls):
        return False
    
    def walk(self, visit, carry):
        visit(self, carry)
        self.term.walk(visit, carry)
        if self.alternate_expression:
            self.alternate_expression.walk(visit, carry)
    
    @classmethod
    def str_list(cls, expressions):
        return ' '.join([str(expression) for expression in expressions])

class Term:
    def __init__(self, value):
        self.value = value

class Optional(Term):
    def __str__(self):
        return '[' + Expression.str_list(self.value) + ']'
    
    def get_children(self):
        return [Clause(self.value), Trivial()]
    
    @classmethod
    def is_disjunction(cls):
        return True
    
    @classmethod
    def is_leaf(cls):
        return False
    
    def walk(self, visit, carry):
        visit(self, carry)
        for expression in self.value:
            expression.walk(visit, carry)

class Reference(Term):
    def __init__(self, name):
        self.reference = None
        Term.__init__(self, name)
    
    def __str__(self):
        return ":" + self.value
    
    def get_children(self):
        return [self.reference]
    
    @classmethod
    def is_disjunction(cls):
        return False
    
    @classmethod
    def is_leaf(cls):
        return False
        
    def walk(self, visit, carry):
        visit(self, carry)

class Stop(Term):
    def __init__(self):
        Term.__init__(self, None)
    
    def __str__(self):
        return '.'
    
    @classmethod
    def is_leaf(cls):
        return True
    
    def walk(self, visit, carry):
        pass

class String(Term):
    def __init__(self, string):
        Term.__init__(self, string)
    
    def __str__(self):
        return "'" + self.value + "'"
    
    @classmethod
    def is_leaf(cls):
        return True
    
    def walk(self, visit, carry):
        visit(self, carry)

class Trivial(Term):
    def __init__(self):
        Term.__init__(self, None)
    
    def __str__(self):
        return "<True>"
    
    @classmethod
    def is_leaf(cls):
        return True
    
    def walk(self, visit, carry):
        pass

class Variable(Term):
    def __init__(self, name):
        self.name = name
        Term.__init__(self, None)
    
    def __str__(self):
        return "<" + str(self.name) + ": " + str(self.value) + ">"
    
    def bind(self, value, context):
        context[self.name] = self
        if not self.value:
            self.value = value
        return self.value == value
    
    @classmethod
    def is_leaf(cls):
        return True
    
    def walk(self, visit, carry):
        visit(self, carry)