
import unittest

from dmzb.slang.pidgin import model
from dmzb.slang.pidgin import lexer
from dmzb.slang.pidgin.recognizer import Recognizer
from dmzb.slang.pidgin.scanner import Scanner
from dmzb.slang.pidgin.unifier import Unifier

class TestUnifier(unittest.TestCase):
    
    def setUp(self):
        data = '''
            channel: [list] channels.
            channel: describe :channel-spec.
            channel-spec: channel [named] <channel-name>
            channel-spec: channel number <channel-number>
            recording: [list] recordings.
            recording: describe recording [with [program] id] <program-id>.
            recording: [list] recordings marked for delete.
            recording: [list] recordings for :channel-spec.
        '''
        self.predicates = lexer.predicates(data)
    
    def recognize(self, command):
        return Recognizer.recognize(self.predicates, Scanner.scan(command))
    
    def test_example(self):
        data = '''
            channel: [list] channels.
            channel: describe :channel-spec.
            channel-spec: channel [named] <channel-name>
            channel-spec: channel number <channel-number>
            recording: [list] recordings.
            recording: describe recording [with [program] id] <program-id>.
            recording: [list] recordings marked for delete.
            recording: [list] recordings for :channel-spec.
        '''
        module = lexer.predicates(data)
        command = "list recordings for channel Disney"
        unificand = Scanner.scan(command)
        unifier = Recognizer.recognize(module, unificand)
        self.assertEqual(unifier.predicate.name, 'recording', 'Not a recording command')
        self.assertTrue('channel-name' in unifier.variables, 'No value for <channel-name>')
        self.assertEqual(unifier.variables['channel-name'], 'Disney', 'Wrong channel name')

    def test_graph(self):
        self.assertIsNotNone(self.predicates, 'No predicates')
        self.assertIsInstance(self.predicates, model.Module, 'Module is not a Module')
    
    def test_command(self):
        command = Scanner.scan("list channels")
        self.assertIsNotNone(command, "No data from scanner")
    
    def test_unify(self):
        unifier = self.recognize("list channels")
        self.assertIsNotNone(unifier, "No frame")
        self.assertIsInstance(unifier, Unifier, "Unify result is not Frame")
    
    def test_string(self):
        string = str(self.recognize("list channels"))
        self.assertIsNotNone(string, "No string representation")
    
    def test_optionals(self):
        self.assertIsNotNone(self.recognize("describe recording 12"))
        self.assertIsNotNone(self.recognize("describe recording with program id 12"))
        self.assertIsNotNone(self.recognize("describe recording with id 12"))
        
    def test_reference(self):
        self.assertIsNotNone(self.recognize("describe channel Disney"))
        self.assertIsNotNone(self.recognize("describe channel named Disney"))
        self.assertIsNotNone(self.recognize("describe channel number 912"))
    
    def test_has_subsequence(self):
        unifier = self.recognize("describe recording with program id 15")
        self.assertIsNotNone(unifier, "No unifier for recording")
        self.assertTrue(unifier.has_subsequence(["with", "program", "id"]), "Failed to find subsequence")
        self.assertFalse(unifier.has_subsequence(["pork", "pie", "hat"]), "Confabulated sequence")
    
    def test_variables(self):
        unifier = self.recognize("describe recording with program id 15")
        self.assertIsNotNone(unifier, "No unifier for recording")
        self.assertTrue("program-id" in unifier.variables, "No program ID in variables")
        self.assertEqual(unifier.variables['program-id'], '15', "Wrong value for program-id")
        unifier = self.recognize("describe channel named Disney")
        self.assertIsNotNone(unifier, "No unifier for channel name")
        self.assertTrue("channel-name" in unifier.variables, "No channel-name in variables")
        self.assertEqual(unifier.variables['channel-name'], 'Disney', "Wrong value for channel-name")
        unifier = self.recognize("describe channel number 912")
        self.assertIsNotNone(unifier, "No unifier for channel number")
        self.assertTrue("channel-number" in unifier.variables, "No channel-number in variables")
        self.assertEqual(unifier.variables['channel-number'], '912', "Wrong value for channel-number")
    
if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
